(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_env.
Require Import Derived_allsT.


(*************************************************************************)
(* N-ary existentials *)
(*************************************************************************)

(* ?a1:k1,...,aN:kN.t   :=   ?a1:k1....?aN:kN.t *)
Fixpoint exsT ks t {struct ks} : typ :=
  match ks with
  | nil => t
  | k::ks' => exT k (exsT ks' t)
  end.

(* (t1,...,tN, e) as ?a1:k1,...,aN:kN.t   :=
   (t1, ... (tN, e) as ?aN:kN.t[t1/aN,...]...) as ?a1:k1....?aN:kN.t *)
Fixpoint packsE ts e t {struct ts} : exp :=
  match ts with
  | nil => e
  | t'::ts' =>
    match t with
    | exT k t'' => packE t' (packsE ts' e (openAT t' t'')) t
    | _ => e
    end
  end.

(* unpack (a1 : k1,...,aN : kN, xN : t1) = e1 in e2   :=
   unpack (a1, x1) = e1 in ... unpack (aN, xN) = x[N-1] in e2 *)
Fixpoint unpacksE' e1 e2 {struct e2} : exp :=
  match e2 with
  | genE k e2' => unpackE e1 (unpacksE' 0 e2')
  | _ => appE e2 e1
  end.
(*
  match ks with
  | nil => letE t1 e1 e2
  | k::ks' => unpackE e1 (unpacksE ks' t1 0 (shiftXE 1 Up 1 e2))
  end.
*)

Notation "'unpacksE' ks t1 e1 e2" :=
  (unpacksE' e1 (gensE ks (lamE t1 e2)))
  (at level 1, ks at level 0, t1 at level 0, e1 at level 0, e2 at level 0).

(*
Lemma varsAT_exs : forall ks t,
  varsAT (exsT ks t) = varsAT t.
Proof.
  intros ks t.
  unfold exsT.
  induction ks.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      unfold varsAT. fold varsAT.
      apply IHks.
Qed.
*)


(*************************************************************************)
(* Substitution on types *)
(*************************************************************************)

Lemma substAT_exs : forall a t' ks t,
  substAT a t' (exsT ks t) = exsT ks (substAT (shifts (length ks) a) t' t).
Proof.
  intros a t' ks t.
  generalize dependent a. generalize dependent t.
  induction ks; intros t a'; simpl exsT; simpl substAT.
    Case "ks = nil".
      reflexivity.
    Case "ks = k::ks'".
      rewrite IHks. rewrite shift_shifts. rewrite shifts_shift. reflexivity.
Qed.

Lemma openAT_exs : forall t' ks t,
  openAT t' (exsT ks t) = exsT ks (substAT (length ks) t' t).
Proof.
  intros t' ks t. unfold openAT.
  rewrite substAT_exs. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.


Lemma substsAT_exs : forall ks as' ts t,
  substsAT as' ts (exsT ks t) = exsT ks (substsAT (map (shifts (length ks)) as') ts t).
Proof.
  induction ks; intros as' ts t; simpl exsT.
    rewrite map_id. reflexivity. 
    rewrite substsAT_ex. 
      rewrite IHks.
      rewrite map_map. do 3 f_equal.
      induction as'; simpl map.
        reflexivity.
        rewrite shift_shifts. rewrite shifts_shift. rewrite IHas'. reflexivity.
Qed.

Lemma opensAT_exs : forall ts ks t,
  opensAT ts (exsT ks t) = exsT ks (substsAT (map (shifts (length ks)) (bnds (length ts))) ts t).
Proof.
  intros ts ks t.
  rewrite unfold_opensAT. rewrite substsAT_exs. reflexivity.
Qed.


(*************************************************************************)
(* Local closure on types *)
(*************************************************************************)

Lemma lclosedT_exs : forall L ks t,
  (forall as', fresh L (length ks) as' -> lclosedT (opensAT (freeTs as') t)) ->
  lclosedT (exsT ks t).
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t Ht; simpl exsT.
    replace t with (opensAT (freeTs nil) t).
      apply Ht. simpl fresh. trivial.
      unfold freeTs. simpl map. simpl opensAT. reflexivity.
    apply lclosedT_ex with (L := L). intros a0 Fr.
      rewrite openAT_exs. apply IHks with (L := L \u {{a0}}). intros as' Frs.
        replace (length ks) with (length (freeTs as')).
          rewrite <- opensAT_cons.
            rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
            apply lclosedT_var.
            apply lclosedTs_vars. rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

(*
Lemma lclosedT_exs_inversion : forall ks t,
  lclosedT (exsT ks t) ->
  exists L,
  (forall as', fresh L (length ks) as' -> lclosedT (opensAT (freeTs as') t)).
Proof.
*)


(*************************************************************************)
(* Substitution on terms *)
(*************************************************************************)

Lemma substAE_packs_var : forall ts t' (a : var) e ks t,
  length ts = length ks ->
  lclosedT t' ->
  substAE a t' (packsE ts e (exsT ks t)) =
  packsE (map (substAT a t') ts) (substAE a t' e) (substAT a t' (exsT ks t)).
Proof.
  induction ts; intros t' a' e ks t Len Ht'; destruct ks; simpl exsT.
    Case "ts = nil, ks = nil".
      reflexivity.
    Case "ts = nil, ks <> nil".
      reflexivity.
    Case "ts <> nil, ks = nil".
      simpl length in Len. inversion Len.
    Case "ts = t::ts', ks = k::ks'l".
      simpl length in Len. inversion Len.
      simpl map. simpl substAT. rewrite substAT_exs.
      simpl packsE.
      simpl substAE. rewrite substAT_exs. f_equal.
      rewrite openAT_exs.
      rewrite IHts. 2: apply H0. 2: apply Ht'.
      f_equal. unfold openAT.
      rewrite <- substAT_exs.
      replace (bnd (length ks)) with (shifts (length ks) 0). 2: rewrite shifts_bnd. 2: f_equal. 2: omega.
      rewrite <- substAT_exs.
      apply substAT_substAT_dep.
        intro Eq. inversion Eq.
        apply Ht'.
Qed.

Lemma substAE_packs : forall ts t' a e ks t,
  length ts = length ks ->
  lclosedT t' ->
  lclosedTs ts ->
  lclosedT (exsT ks t) ->
  substAE a t' (packsE ts e (exsT ks t)) =
  packsE (map (substAT a t') ts) (substAE a t' e) (substAT a t' (exsT ks t)).
Proof.
  induction ts; intros t' a' e ks t Len Ht' Hts Ht; destruct ks; simpl exsT; simpl packsE.
    Case "ts = nil, ks = nil".
      reflexivity.
    Case "ts = nil, ks <> nil".
      reflexivity.
    Case "ts <> nil, ks = nil".
      simpl length in Len. inversion Len.
    Case "ts = t::ts', ks = k::ks'l".
      inversion Ht; subst.
      assert (Hexs' : lclosedT (openAT a (exsT ks t))).
        pick_fresh_gen (L \u varsAT (exsT ks t)) aa.
        rewrite openAT_as_substAT with (a := aa).
          apply lclosedT_substAT.
            eapply lclosedTs_head. apply Hts.
            apply H0. notin_solve.
          notin_solve.
      simpl substAE; simpl substAT; simpl. f_equal.
        rewrite openAT_exs.
        rewrite IHts. f_equal.
          rewrite <- openAT_exs.
          destruct a'; simpl shift.
            unfold openAT. rewrite <- substAT_substAT_dep.
              reflexivity.
              intro. inversion H.
              unfold indepAT. apply Ht'.
            replace (substAT n t' a) with a.
              rewrite openAT_substAT.
                rewrite substAT_lclosed. rewrite substAT_lclosed.
                  reflexivity.
                  apply Hexs'. apply Hexs'.
                intro Eq. inversion Eq.
                unfold indepAT. eapply lclosedTs_head. apply Hts.
                apply Ht'.
              symmetry. apply substAT_lclosed. eapply lclosedTs_head. apply Hts.
          simpl length in Len. inversion Len. reflexivity.
          apply Ht'.
          eapply lclosedTs_tail. apply Hts.
          replace (bnd (length ks)) with (shifts (length ks) 0).
            rewrite <- substAT_exs. apply Hexs'.
            rewrite shifts_bnd. f_equal. omega.
Qed.

Corollary openAE_packs : forall t' ts e ks t,
  length ts = length ks ->
  lclosedT t' ->
  lclosedTs ts ->
  lclosedT (exsT ks t) ->
  openAE t' (packsE ts e (exsT ks t)) =
  packsE (map (openAT t') ts) (openAE t' e) (openAT t' (exsT ks t)).
Proof.
  intros t' ts e ks t Len Ht' Hts Ht. unfold openAE.
  rewrite substAE_packs. reflexivity. apply Len. apply Ht'. apply Hts. apply Ht.
Qed.

Lemma substsAE_packs : forall as' ts' ts e ks t,
  length ts = length ks ->
  lclosedTs ts' ->
  lclosedTs ts ->
  lclosedT (exsT ks t) ->
  substsAE as' ts' (packsE ts e (exsT ks t)) =
  packsE (map (substsAT as' ts') ts) (substsAE as' ts' e) (substsAT as' ts' (exsT ks t)).
Proof.
  induction as'; intros ts' ts e ks t Len Hts' Hts Ht; simpl substsAE; simpl substsAT.
    rewrite map_id. reflexivity.
    destruct ts'.
      rewrite map_id. reflexivity.
      rewrite IHas'.
        rewrite substsAT_exs. rewrite substAE_packs.
          rewrite map_map. reflexivity.
          rewrite map_length. apply Len.
          eapply lclosedTs_head. apply Hts'.
          apply lclosedTs_substsAT. eapply lclosedTs_tail. apply Hts'. apply Hts.
          rewrite <- substsAT_exs. apply lclosedT_substsAT.
            eapply lclosedTs_tail. apply Hts'.
            apply Ht.
        apply Len.
        eapply lclosedTs_tail. apply Hts'.
        apply Hts.
        apply Ht.
Qed.

Corollary opensAE_packs : forall ts' ts e ks t,
  length ts = length ks ->
  lclosedTs ts' ->
  lclosedTs ts ->
  lclosedT (exsT ks t) ->
  opensAE ts' (packsE ts e (exsT ks t)) =
  packsE (map (opensAT ts') ts) (opensAE ts' e) (opensAT ts' (exsT ks t)).
Proof.
  intros ts' ts e ks t Len Hts' Hts Ht.
  rewrite unfold_opensAE.
  rewrite substsAE_packs.
  rewrite <- unfold_opensAE. rewrite <- unfold_opensAT. rewrite <- unfold_opensATs. reflexivity.
  apply Len. apply Hts'. apply Hts. apply Ht.
Qed.

Lemma substXE_packs : forall x e' ts e ks t,
  length ts = length ks ->
  substXE x e' (packsE ts e (exsT ks t)) = packsE ts (substXE x e' e) (exsT ks t).
Proof.
  intros x e'.
  induction ts; intros e ks t Len; destruct ks; simpl packsE; simpl substXE.
    Case "ts = nil, ks = nil".
      reflexivity.
    Case "ts = nil, ks <> nil".
      simpl length in Len. inversion Len.
    Case "ts <> nil, ks = nil".
      simpl length in Len. inversion Len.
    Case "ts = t::ts', ks = k::ks'".
      rewrite openAT_exs.
      rewrite IHts.
        reflexivity.
        simpl length in Len. inversion Len. reflexivity.
Qed.

Corollary openXE_packs : forall e' ts e ks t,
  length ts = length ks ->
  openXE e' (packsE ts e (exsT ks t)) = packsE ts (openXE e' e) (exsT ks t).
Proof.
  intros e' ts e ks t Len. unfold openXE. apply substXE_packs. apply Len.
Qed.


Lemma substAE_unpacks : forall a t ks t1 e1 e2,
  substAE a t (unpacksE ks t1 e1 e2) =
  unpacksE ks (substAT (shifts (length ks) a) t t1) (substAE a t e1) (substAE (shifts (length ks) a) t e2).
Proof.
  intros a t ks t1 e1 e2.
  generalize dependent a. generalize dependent e1. generalize dependent e2.
  induction ks; intros e2 e1 a'; simpl unpacksE'.
    Case "ks = nil".
      simpl substAE. reflexivity.
    Case "ks = k::ks'".
      simpl substAE.
      f_equal.
      rewrite IHks.
        simpl substAE. rewrite shifts_shift. rewrite shift_shifts. reflexivity.
Qed.
(*
Lemma substAE_unpacks : forall a t ks t1 e1 e2,
  lclosedT t ->
  lclosedE e2 ->
  substAE a t (unpacksE ks t1 e1 e2) =
  unpacksE ks (substAT (shifts (length ks) a) t t1) (substAE a t e1) (substAE (shifts (length ks) a) t e2).
Proof.
  intros a t ks t1 e1 e2.
  generalize dependent a. generalize dependent e1. generalize dependent e2.
  induction ks; intros e2 e1 a' Ht He2; simpl unpacksE.
    Case "ks = nil".
      apply substAE_let.
    Case "ks = k::ks'".
      unfold letsE, appsE, lamsE.
      simpl substAE.
      f_equal.
      rewrite IHks.
        simpl substAE. simpl substAE.
          rewrite shifts_shift. rewrite shift_shifts.
          destruct a'.
            rewrite shifts_free. rewrite shiftXE_lclosed. rewrite shiftXE_lclosed.
              reflexivity.
              apply lclosedE_substAE. apply Ht.
              apply He2.
              apply He2.
            rewrite shifts_bnd. rewrite shiftXE_lclosed. rewrite shiftXE_lclosed.
              reflexivity.
              apply lclosedE_substAE. apply Ht.
              apply He2.
              apply He2.
        apply Ht.
        rewrite shiftXE_lclosed. apply He2. apply He2.
Qed.
*)

Corollary openAE_unpacks : forall t ks t1 e1 e2,
  openAE t (unpacksE ks t1 e1 e2) =
  unpacksE ks (substAT (length ks) t t1) (openAE t e1) (substAE (length ks) t e2).
Proof.
  intros t ks t1 e1 e2. unfold openAE, openAT.
  replace (bnd (length ks)) with (shifts (length ks) 0). 
    apply substAE_unpacks.
    replace (bnd (length ks)) with (bnd (length ks + 0)). apply shifts_bnd. f_equal. omega.
Qed.

Lemma substsAE_unpacks : forall ks as' ts t1 e1 e2,
  substsAE as' ts (unpacksE ks t1 e1 e2) =
  unpacksE ks (substsAT (map (shifts (length ks)) as') ts t1) (substsAE as' ts e1) (substsAE (map (shifts (length ks)) as') ts e2).
Proof.
  induction ks; intros as' ts t1 e1 e2; simpl unpacksE'.
    rewrite map_id. rewrite substsAE_app. rewrite substsAE_lam. reflexivity.
    rewrite substsAE_unpack.
      rewrite IHks.
      rewrite map_map. do 3 f_equal.
      generalize dependent ts. induction as'; intros ts; simpl map.
        reflexivity.
        destruct ts. reflexivity. simpl substsAE. rewrite IHas'. reflexivity.
      do 2 f_equal.
        induction as'; simpl map.
          reflexivity.
          rewrite shifts_shift. rewrite shift_shifts. rewrite IHas'. reflexivity.
        induction as'; simpl map.
          reflexivity.
          rewrite shifts_shift. rewrite shift_shifts. rewrite IHas'. reflexivity.
Qed.

Corollary opensAE_unpacks : forall ts ks t1 e1 e2,
  opensAE ts (unpacksE ks t1 e1 e2) =
  unpacksE ks (substsAT (map (shifts (length ks)) (bnds (length ts))) ts t1) (opensAE ts e1) (substsAE (map (shifts (length ks)) (bnds (length ts))) ts e2).
Proof.
  intros ts ks t1 e1 e2.
  do 2 rewrite unfold_opensAE. rewrite substsAE_unpacks. reflexivity.
Qed.

Lemma substXE_unpacks : forall x e' ks t1 e1 e2,
  substXE x e' (unpacksE ks t1 e1 e2) =
  unpacksE ks t1 (substXE x e' e1) (substXE (shift (shifts (length ks) x)) e' e2).
Proof.
  intros x e' ks t1.
  generalize dependent x.
  induction ks; intros x e1 e2; simpl unpacksE'.
    Case "ks = nil".
      simpl substXE. reflexivity.
    Case "ks = k::ks'".
      simpl substXE.
      f_equal.
      rewrite IHks.
      f_equal.
        apply substXE_var_ne. intros Eq. symmetry in Eq. destruct (shift_nonzero _ Eq).
        rewrite shifts_shift. rewrite shift_shifts. rewrite shift_shifts. rewrite shift_shifts.
        reflexivity.
Qed.

Corollary openXE_unpacks : forall e' ks t1 e1 e2,
  openXE e' (unpacksE ks t1 e1 e2) = unpacksE ks t1 (openXE e' e1) (substXE (S (length ks)) e' e2).
Proof.
  intros e' ks t1 e1 e2. unfold openXE.
  replace (bnd (S (length ks))) with (shift (shifts (length ks) 0)).
    apply substXE_unpacks.
    rewrite shift_shifts. rewrite shifts_bnd. f_equal. omega.
Qed.

(*
Lemma substXE_unpacks : forall x e' ks t1 e1 e2,
  lclosedE e' ->
  lclosedE e2 ->
  substXE x e' (unpacksE ks t1 e1 e2) = unpacksE ks t1 (substXE x e' e1) (substXE (shift x) e' e2).
Proof.
  intros x e' ks t1 e1 e2.
  generalize dependent x. generalize dependent e1. generalize dependent e2.
  induction ks; intros e2 e1 x He' He2; simpl unpacksE.
    Case "ks = nil".
      apply substXE_let.
    Case "ks = k::ks'".
      unfold letsE, appsE, lamsE.
      simpl substXE.
      f_equal.
      rewrite IHks.
        f_equal.
          apply substXE_var_ne. intros Eq. symmetry in Eq. destruct (shift_nonzero _ Eq).
          rewrite shiftXE_lclosed. rewrite shiftXE_lclosed.
            destruct x; simpl shift; simpl.
              reflexivity.
              rewrite substXE_lclosed. rewrite substXE_lclosed.
                reflexivity.
                apply He2. apply He2.
            apply lclosedE_substXE. apply He'. apply He2.
            apply He2. apply He'.
        rewrite shiftXE_lclosed. apply He2. apply He2.
Qed.
*)


(*************************************************************************)
(* Local closure on terms *)
(*************************************************************************)

Lemma lclosedE_packs : forall ts e ks t,
  length ts = length ks ->
  lclosedTs ts ->
  lclosedE e ->
  lclosedT (exsT ks t) ->
  lclosedE (packsE ts e (exsT ks t)).
Proof.
  induction ts; intros e ks t Len Hts He Ht; destruct ks; simpl packsE.
    Case "ts = nil, ks = nil".
      apply He.
    Case "ts = nil, ks <> nil".
      simpl length in Len. inversion Len.
    Case "ts <> nil, ks = nil".
      simpl length in Len. inversion Len.
    Case "ts = t::ts', ks = k::ks'".
      inversion Ht; subst.
      assert (Hexs' : lclosedT (openAT a (exsT ks t))).
        pick_fresh_gen (L \u varsAT (exsT ks t)) aa.
        rewrite openAT_as_substAT with (a := aa).
          apply lclosedT_substAT.
            eapply lclosedTs_head. apply Hts.
            apply H0. notin_solve.
          notin_solve.
      apply lclosedE_pack.
        eapply lclosedTs_head. apply Hts.
        rewrite openAT_exs. apply IHts.
          simpl length in Len. inversion Len. reflexivity.
          eapply lclosedTs_tail. apply Hts.
          apply He.
          replace (bnd (length ks)) with (shifts (length ks) 0).
            rewrite <- substAT_exs. apply Hexs'.
            rewrite shifts_bnd. f_equal. omega.
        eapply lclosedT_ex. intros a0 Fr. apply H0. apply Fr.
Qed.

Lemma lclosedE_unpacks : forall ks t1 e1 e2,
  lclosedE e1 ->
  lclosedE (gensE ks (lamE t1 e2)) ->
  lclosedE (unpacksE ks t1 e1 e2).
Proof.
  induction ks; intros t1 e1 e2 He1 He2; simpl unpacksE'.
    Case "ks = nil".
      simpl gensE in He2. inversion He2; subst.
      eapply lclosedE_app. eapply lclosedE_lam.
        apply H1.
        apply H2.
        apply He1.
    Case "ks = k::ks'".
      simpl gensE in He2. inversion He2; subst.
      eapply lclosedE_unpack.
        apply He1.
        intros a0 x Fra Frx.
          rewrite openAE_unpacks. rewrite openXE_unpacks.
          unfold openAE. simpl substAE.
          unfold openXE. simpl substXE.
          apply IHks.
            apply lclosedE_var.
            rewrite substXE_substAE.
              assert (Aux : forall a t' t e, lamE (substAT a t' t) (substAE a t' e) = substAE a t' (lamE t e)).
                intros. simpl substAE. reflexivity.
              rewrite Aux.
              replace (bnd (length ks)) with (shifts (length ks) 0). rewrite <- substAE_gens.
              assert (Aux2 : forall n e' t e, lamE t (substXE (S n) e' e) = substXE n e' (lamE t e)).
                intros. simpl substXE. reflexivity.
              rewrite Aux2. rewrite <- substXE_gens.
              rewrite substAE_substXE.
              unfold openAE in H0.
              rewrite substXE_lclosed.
                apply H0. apply Fra.
                apply H0. apply Fra.
              unfold indepAE. apply lclosedE_var.
              rewrite shifts_bnd. f_equal. omega.
              unfold indepAE. apply lclosedE_var.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofT_exs : forall L D ks t,
  okD D ->
  (forall as', fresh L (length ks) as' ->
   ofT (D & as'~~ks) (opensAT (freeTs as') t) omK) ->
  ofT D (exsT ks t) omK.
Proof.
  intros L D ks t.
  generalize dependent t.
  generalize dependent D. generalize dependent L.
  induction ks; intros L D t HD Ht; simpl exsT in *.
    Case "ks = nil".
      assert (Ht' : ofT (D & nil~~nil) (opensAT (freeTs nil) t) omK).
        apply Ht. simpl fresh. trivial.
      apply Ht'.
    Case "ks = k :: ks'".
      apply ofT_ex with (L := L \u dom D).
        intros a0 Fr. rewrite openAT_exs.
          apply IHks with (L := L \u {{a0}}).
            apply okD_var. apply HD. notin_solve.
            intros as' Frs.
              rewrite concat_assoc. rewrite <- concat_multi_cons.
              replace (length ks) with (length as').
                rewrite <- opensAT_cons_vars. apply Ht.
                  simpl fresh. split. notin_solve. apply Frs.
                symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma ofT_exs_inversion : forall D ks t,
  ofT D (exsT ks t) omK ->
  exists L,
  (forall as', fresh L (length ks) as' ->
   ofT (D & as'~~ks) (opensAT (freeTs as') t) omK).
Proof.
  intros D ks; generalize D.
  clear D.
  induction ks; intros D t.
  Case "ks = nil".
     intros. 
     exists {}.
     intros.
     simpl length in H0.
     assert (RW := fresh_length _  _ _ H0).
     symmetry in RW.
     assert (RWas' := length_zero_is_nil _ _ RW).
     rewrite RWas'.
     apply H.
  Case "ks = a::ks".
     intro H.
     simpl allsT in  H.
     inversion H.
     pick_fresh_gen (L \u varsAT t ) a0.
     assert (FrL : a0 \notin L).
      notin_solve.
     assert (H2a := H2 a0 FrL).
     rewrite openAT_exs in H2a.
     assert (IHH2a := IHks _ _ H2a).
     destruct IHH2a as [L1 H4].
     exists (L1 \u {{a0}} \u dom(D)).
     intros.
     assert (RW1 := fresh_length _ _ _ H5).
     simpl length in RW1.
     symmetry in RW1.
     assert (H6 := length_succ_is_cons _ _ _ RW1).
     destruct H6 as [b [bs RWas']].
     rewrite RWas'. 
     rewrite concat_multi_cons.     
     unfold freeTs. simpl.
     fold (freeTs bs).
     rewrite substAT_opensAT.
     2: omega. 
     2: apply lclosedT_var. 
     2: apply lclosedTs_vars.
     assert (RWlen_free_bs : length (freeTs bs) = length ks).
      rewrite RWas' in RW1.
      injection RW1.
      intros RWlenbs.
      unfold freeTs.
      rewrite map_length.
      rewrite map_length.
      exact RWlenbs.
     assert (H6 : ofT (D & a0 ~ a & bs ~~ ks) (opensAT (freeTs bs) (substAT (length ks) a0 t)) omK).
       apply H4.
       simpl in H5.
       rewrite RWas' in H5.
       simpl fresh in H5.
       destruct H5 as [_ H6].
       fresh_solve.
     assert (H7 : ofT (D & b ~ a & bs ~~ ks) (substAT a0 b (opensAT (freeTs bs) (substAT (length ks) a0 t))) omK).
      eapply (substitutabilityAT ).
      rewrite concat_assoc.
      eapply weakeningDT.
      rewrite <- concat_assoc.
      apply H6.
      (*   okD (D & b ~ a & (a0 ~ a & bs ~~ ks)), plausible *)
      rewrite <- concat_assoc.
      apply okD_vars.
       apply okD_var.
         apply okD_var.
         apply (validityDT _ _ _ H).
          rewrite RWas' in H5.
          fresh_solve.
          simpl.
          apply notin_union_l.
          apply notin_singleton_swap.
          rewrite RWas' in H5.
          fresh_solve.
          assert ( H7 := validityDT _ _ _ H2a).
          inversion H7; auto.
          simpl.
          rewrite RWas' in H5.
          simpl fresh in H5.
          destruct H5 as [_ H7].
          fresh_solve.
          (* ofT (D & b ~ a & bs ~~ ks) b a, clearly *)
         rewrite concat_assoc.
         rewrite <- concat_multi_cons.
         apply ofT_vars.
          rewrite RWas' in H5.
          apply okD_vars.
           apply (validityDT _ _ _ H).
           rewrite <- union_assoc in H5.
           assert (H7 := fresh_union_r _ _ _ _ H5).
           destruct H7 as [_ H8].
           assert (H9 := fresh_union_r _ _ _ _ H8).
           firstorder.
        simpl.
        apply in_or_app.
        right.
        auto.
     rewrite <- concat_assoc.
     rewrite <- substAT_opensAT in H7.
     rewrite <- substAT_bnd_as_substAT_free in H7.
     rewrite <- substAT_opensAT.    
     rewrite RWlen_free_bs.
     apply H7.
     omega.
     apply lclosedT_var.
     apply lclosedTs_vars.
      eapply notin_varsAT_opensAT.
       rewrite RWas' in H5.
       simpl fresh in H5.
       destruct H5 as [_ H8].
       assert (H9 := fresh_resize _ _ _ H8).
       assert (H10 := fresh_union_r _ _ _ _ H9).
       destruct H10 as [H11 _].
       assert (H12 := fresh_union_r _ _ _ _ H11).
       destruct H12 as [H13 _].
         eapply H13.
      fresh_solve. (*  a0 \notin varsAT t, by our choice of a0! *) 
     rewrite RWlen_free_bs.
     omega.
     apply lclosedT_var.
     apply lclosedTs_vars.
Qed.


Lemma eqT_exs : forall L ks t t',
  (forall as', fresh L (length ks) as' -> eqT (opensAT (freeTs as') t) (opensAT (freeTs as') t')) ->
  eqT (exsT ks t) (exsT ks t').
Proof.
  intros L ks. generalize dependent L.
  induction ks; intros L t t' Ht; simpl exsT.
    Case "ks = nil".
      replace t with (opensAT (freeTs nil) t). replace t' with (opensAT (freeTs nil) t').
        apply Ht. unfold length, fresh. trivial.
        unfold freeTs, opensAT. simpl. reflexivity.
        unfold freeTs, opensAT. simpl. reflexivity.
    Case "ts = t::ts".
      eapply eqT_ex. intros a0 Fr.
      rewrite openAT_exs. rewrite openAT_exs.
      apply IHks with (L := L \u {{a0}}). intros as' Frs.
      replace (length ks) with (length (freeTs as')).
      rewrite <- opensAT_cons. rewrite <- opensAT_cons.
        rewrite <- freeTs_cons. apply Ht. simpl fresh. split. apply Fr. apply Frs.
        apply lclosedT_var. apply lclosedTs_vars.
        apply lclosedT_var. apply lclosedTs_vars.
        rewrite freeTs_length. symmetry. eapply fresh_length. apply Frs.
Qed.


Lemma ofE_packs : forall D G ts e ks t,
  length ts = length ks ->
  (forall t k, In (t, k) (combine ts ks) -> ofT D t k) ->
  ofE D G e (opensAT ts t) ->
  ofT D (exsT ks t) omK ->
  ofE D G (packsE ts e (exsT ks t)) (exsT ks t).
Proof.
  intros D G.
  induction ts; intros e ks t Len Hts He Ht; destruct ks; simpl packsE in *; simpl exsT in *.
    Case "ts = nil, ks = nil".
      apply He.
    Case "ts = nil, ks <> nil".
      simpl length in Len. inversion Len.
    Case "ts <> nil, ks = nil".
      simpl length in Len. inversion Len.
    Case "ts = t::ts', ks = k::ks'".
      inversion Ht; subst.
      assert (Hexs' : ofT D (openAT a (exsT ks t)) omK).
        pick_fresh_gen (L \u varsAT (exsT ks t)) aa.
        rewrite openAT_as_substAT with (a := aa).
          eapply substitutabilityAT_head.
            apply H1. notin_solve.
            apply Hts. simpl combine. apply in_eq.
          notin_solve.
      apply ofE_pack.
        apply Hts. simpl combine. apply in_eq.
        rewrite openAT_exs. apply IHts.
          simpl length in Len. inversion Len. reflexivity.
          intros t0 k0 Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
          replace (length ks) with (length ts).
            rewrite <- opensAT_cons.
              apply He.
              eapply regularityT. apply Hts. simpl combine. apply in_eq.
              intros t0 Elem.
                assert (Elem' : In t0 (a::ts)). apply in_cons. apply Elem.
                specialize (in_combine_l_inv typ knd (a::ts) (k::ks) t0 Len Elem').
                  intro H. destruct H.
                eapply regularityT. apply Hts. apply H.
            simpl length in Len. inversion Len. reflexivity.
          rewrite <- openAT_exs. apply Hexs'.
        apply Ht.
Qed.

Lemma ofE_unpacks : forall ks L D G e1 t1 e2 t2,
  ofE D G e1 (exsT ks t1) ->
  (forall as' x, fresh L (length ks) as' -> x \notin L ->
     ofE (D & as'~~ks) (G & x~(opensAT (freeTs as') t1))
       (openXE x (opensAE (freeTs as') e2)) t2) ->
  ofT D t2 omK ->
  ofE D G (unpacksE ks t1 e1 e2) t2.
Proof.
  induction ks; intros L D G e1 t1 e2 t2 He1 He2 Ht2; simpl unpacksE'; simpl exsT in *.
    Case "ks = nil".
      eapply ofE_app.
        eapply ofE_lam.
          eapply validityTE. apply He1.
          intros x Fr. apply (He2 nil x). simpl. trivial. apply Fr.
        apply He1.
    Case "ks = k::ks'".
      assert (HD : okD D). eapply validityDE. apply He1.
      assert (HG : okG D G). eapply validityGE. apply He1.
      assert (Ht1' : ofT D (exT a (exsT ks t1)) omK). eapply validityTE. apply He1.
      inversion Ht1'; subst.
      assert (Ht1'' : ofT D (exsT (a::ks) t1) omK). apply Ht1'.
      destruct (ofT_exs_inversion _ _ _ Ht1'') as [L' Ht1'''].
      eapply ofE_unpack with (L := L \u L0 \u L' \u dom D \u dom G).
        apply He1.
        intros a0 x0 Fra0 Frx0.
          assert (HD' : okD (D & a0~a)). apply okD_var. apply HD. notin_solve.
          assert (HG' : okG (D & a0~a) (G & x0 ~ openAT a0 (exsT ks t1))).
            apply okG_var.
              eapply weakeningDG_head. apply HG. apply HD'.
              notin_solve.
              apply H1. notin_solve.
          rewrite openAE_unpacks. rewrite openXE_unpacks.
          apply IHks with (L := L \u L0 \u L' \u dom D \u {{a0}} \u dom G \u {{x0}}).
            unfold openAE. simpl substAE. unfold openXE. simpl substXE.
              unfold openAT in *. rewrite substAT_exs in *. rewrite shifts_bnd in *. rewrite plus_zero in *.
              apply ofE_var.
              apply HG'.
              apply binds_head.
              intros as' x Fras Frx.
                assert (Fras' : fresh L (length (a::ks)) (a0::as')).
                  simpl length. simpl fresh. split.
                    notin_solve.
                    apply fresh_union_l.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                        apply Fras.
                      eapply proj2. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                        apply Fras.
                assert (Fras'' : fresh L' (length (a::ks)) (a0::as')).
                  simpl length. simpl fresh. split.
                    notin_solve.
                    apply fresh_union_l.
                      eapply proj2. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                        apply Fras.
                      eapply proj2. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                      eapply proj1. apply fresh_union_r.
                        apply Fras.
                rewrite concat_assoc. rewrite <- concat_multi_cons.
                replace (length ks) with (length (freeTs as')).
                rewrite <- opensAT_cons.
                rewrite <- substXE_opensAE.
                rewrite openXE_substXE.
                rewrite <- opensAE_cons.
                replace (varT a0 :: freeTs as') with (freeTs (a0::as')).
                rewrite substXE_lclosed.
                apply weakeningGE.
                apply He2.
                  apply Fras'.
                  notin_solve.
                  eapply okG_var.
                    rewrite concat_multi_cons. rewrite <- concat_assoc.
                      apply weakeningDG_head.
                        apply HG'.
                        apply okD_vars.
                          apply HD'.
                          simpl dom. apply fresh_union_l.
                            eapply proj2. apply fresh_union_r.
                            eapply proj1. apply fresh_union_r.
                            eapply proj1. apply fresh_union_r.
                              apply Fras.
                            eapply proj2. apply fresh_union_r.
                            eapply proj1. apply fresh_union_r.
                            eapply proj1. apply fresh_union_r.
                            eapply proj1. apply fresh_union_r.
                              apply Fras.
                        simpl dom. notin_solve.
                    apply Ht1'''. apply Fras''.
                  eapply regularityE. apply He2.
                    apply Fras'.
                    notin_solve.
                  unfold freeTs. simpl map. reflexivity.
                  apply lclosedT_var.
                  apply lclosedTs_vars.
                  intro Eq. inversion Eq.
                  unfold indepXE. apply lclosedE_var. 
                  apply lclosedE_var.
                  apply lclosedE_var.
                  apply lclosedT_var.
                  apply lclosedTs_vars.
                  rewrite freeTs_length. symmetry. eapply fresh_length. apply Fras.
            eapply weakeningDT_head. apply Ht2. apply HD'.
        apply Ht2.
Qed.

(*============
Lemma ofT_ex_inversion : forall D k t,
  ofT D (exT k t) omK ->
  exists L,
  (forall a, a \notin L -> ofT (D & a~k) (openAT a t) omK).
Proof.
  intros D k t H. inversion H. exists L. intros. apply H2. apply H4.
Qed.

Lemma ofT_exs_inversion : forall D ks t,
  ofT D (exsT ks t) omK ->
  exists L,
  (forall as', fresh L (length ks) as' ->
   ofT (D & as'~~ks) (opensAT (freeTs as') t) omK).
Proof.
  intros D ks t Ht.
  unfold exsT in Ht.
  generalize dependent t. generalize dependent D.
  induction ks; intros D t Ht.
    Case "ks = nil".
      exists {}.
      intros as' Fr.
      destruct as'.
        SCase "as' = nil".
          unfold rev, combine. fold (empty (A := knd)). rewrite concat_empty.
          unfold freeTs, opensAT, map. apply Ht.
        SCase "as' = a::as".
          unfold length, fresh in Fr. destruct Fr.
    Case "ks = k::ks'".
      inversion Ht; subst.
      exists L.
      intros as' Fr.
      destruct as'.
        SCase "as' = nil".
          unfold length, fresh in Fr. destruct Fr.
        SCase "as' = a::as".
          rewrite concat_multi_cons. rewrite <- concat_assoc.
          fold exsT in *.
          rewrite <- lemma_opensAT_list.
          unfold length, fresh in Fr. destruct Fr.
          replace (length as') with (length ks).
            assert (Ht' : ofT (D & v~a) (exsT ks (substAT (length ks) v t)) omK).
              rewrite <- lemma_openAT_exsT.
                apply H1. notin_solve.
                apply lclosedT_var.
            assert (IH' := IHks (D & v~a) (substAT (length ks) v t) Ht').
            destruct IH' as [L' IH''].
            apply IH''.

(*
          apply IHks.
            apply H3.
            inversion Ht; subst. 
              rewrite <- lemma_openAT_exsT.
                apply H3. apply H.

        intros a0 Fr.
        rewrite lemma_openAT_exsT.
        apply IHks with (L := L \u {{a0}}).
          apply okD_var. apply HD. notin_solve.
          intros as' Frs.
          rewrite concat_assoc. 
          rewrite <- concat_multi_cons.
          replace (opensAT (freeTs as') (t [length ks -AT> a0])) with (opensAT (freeTs (a0::as')) t).
          apply Ht. unfold fresh. simpl. split. notin_solve. apply Frs.
          unfold freeTs. simpl.
          rewrite map_length. rewrite map_length.
          replace (length ks) with (length as').
          reflexivity. symmetry. eapply fresh_length. apply Frs.
          apply lclosedT_var.
Qed.
*)
=======*)


(*************************************************************************)
(* Splitting *)
(*************************************************************************)

Lemma exsT_concat : forall ks1 ks2 t,
  exsT (ks1 ++ ks2) t = exsT ks1 (exsT ks2 t).
Proof.
  induction ks1; intros ks2 t; simpl; try rewrite IHks1; reflexivity.
Qed.

Lemma packsE_concat : forall ts1 ts2 e ks1 ks2 t,
  length ts1 = length ks1 ->
  lclosedTs ts1 ->
  packsE (ts1 ++ ts2) e (exsT (ks1 ++ ks2) t) =
  packsE ts1 (packsE ts2 e (opensAT ts1 (exsT ks2 t))) (exsT (ks1 ++ ks2) t).
Proof.
  induction ts1; intros ts2 e ks1 ks2 t Len Hts1; destruct ks1; simpl length in Len; inversion Len; simpl.
    reflexivity.
    rewrite openAT_exs. rewrite IHts1.
      do 3 f_equal.
        rewrite app_length.
        rewrite substAT_opensAT.
          rewrite substAT_exs.
            do 3 f_equal.
              rewrite shifts_bnd. f_equal. omega.
            omega.
            eapply lclosedTs_head. apply Hts1.
            eapply lclosedTs_tail. apply Hts1.
      apply H0.
      eapply lclosedTs_tail. apply Hts1.
Qed.
