(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Fomega_syntax.
Require Import Fomega_vars.


(*************************************************************************)
(* Kinding *)
(*************************************************************************)

Inductive okD : tenv -> Prop :=
  | okD_empty :
      okD empty
  | okD_var : forall a k D,
      okD D ->
      a \notin dom D ->
      okD (D & a~k)
.

Inductive ofT : tenv -> typ -> knd -> Prop :=
  | ofT_var : forall D (a : var) k,
      okD D ->
      binds a k D ->
      ofT D (varT a) k
  | ofT_unit : forall D,
      okD D ->
      ofT D (unitT) omK
  | ofT_arr : forall D t1 t2,
      ofT D t1 omK ->
      ofT D t2 omK ->
      ofT D (arrT t1 t2) omK
  | ofT_tim : forall D t1 t2,
      ofT D t1 omK ->
      ofT D t2 omK ->
      ofT D (timT t1 t2) omK
  | ofT_plus : forall D t1 t2,
      ofT D t1 omK ->
      ofT D t2 omK ->
      ofT D (plusT t1 t2) omK
  | ofT_rec : forall L D k t,
      (forall a : var, a \notin L -> ofT (D & a~k) (openAT a t) k) ->
      ofT D (recT k t) k
  | ofT_all : forall L D k t,
      (forall a : var, a \notin L -> ofT (D & a~k) (openAT a t) omK) ->
      ofT D (allT k t) omK
  | ofT_ex : forall L D k t,
      (forall a : var, a \notin L -> ofT (D & a~k) (openAT a t) omK) ->
      ofT D (exT k t) omK
  | ofT_triv : forall D,
      okD D ->
      ofT D (trivT) unitK
  | ofT_lam : forall L D k1 t2 k2,
      (forall a : var, a \notin L -> ofT (D & a~k1) (openAT a t2) k2) ->
      ofT D (lamT k1 t2) (arrK k1 k2)
  | ofT_app : forall D t1 t2 k2 k,
      ofT D t1 (arrK k2 k) ->
      ofT D t2 k2 ->
      ofT D (appT t1 t2) k
  | ofT_pair : forall D t1 t2 k1 k2,
      ofT D t1 k1 ->
      ofT D t2 k2 ->
      ofT D (pairT t1 t2) (timK k1 k2)
  | ofT_fst : forall D t k1 k2,
      ofT D t (timK k1 k2) ->
      ofT D (fstT t) k1
  | ofT_snd : forall D t k1 k2,
      ofT D t (timK k1 k2) ->
      ofT D (sndT t) k2
.

Hint Constructors ofT.

(*************************************************************************)
(* Type Equivalence *)
(*************************************************************************)

Inductive eqT : typ -> typ -> Prop :=
  | eqT_var : forall a,
      eqT (varT (free a)) (varT (free a))
  | eqT_unit :
      eqT (unitT) (unitT)
  | eqT_arr : forall t1 t2 t1' t2',
      eqT t1 t1' ->
      eqT t2 t2' ->
      eqT (arrT t1 t2) (arrT t1' t2')
  | eqT_tim : forall t1 t2 t1' t2',
      eqT t1 t1' ->
      eqT t2 t2' ->
      eqT (timT t1 t2) (timT t1' t2')
  | eqT_plus : forall t1 t2 t1' t2',
      eqT t1 t1' ->
      eqT t2 t2' ->
      eqT (plusT t1 t2) (plusT t1' t2')
  | eqT_rec : forall L k t t',
      (forall a : var, a \notin L -> eqT (openAT a t) (openAT a t')) ->
      eqT (recT k t) (recT k t')
  | eqT_all : forall L k t t',
      (forall a : var, a \notin L -> eqT (openAT a t) (openAT a t')) ->
      eqT (allT k t) (allT k t')
  | eqT_ex : forall L k t t',
      (forall a : var, a \notin L -> eqT (openAT a t) (openAT a t')) ->
      eqT (exT k t) (exT k t')
  | eqT_triv :
      eqT (trivT) (trivT)
  | eqT_lam : forall L k t t',
      (forall a : var, a \notin L -> eqT (openAT a t) (openAT a t')) ->
      eqT (lamT k t) (lamT k t')
  | eqT_lam_eta : forall k t,
      lclosedT t ->
      eqT (lamT k (appT t (varT (bnd 0)))) t
  | eqT_app : forall t1 t2 t1' t2',
      eqT t1 t1' ->
      eqT t2 t2' ->
      eqT (appT t1 t2) (appT t1' t2')
  | eqT_app_beta : forall k t1 t2,
(*      lclosedT t1 -> (*BUG!*) *)
      lclosedT (lamT k t1) ->
      lclosedT t2 ->
      eqT (appT (lamT k t1) t2) (openAT t2 t1)
  | eqT_pair : forall t1 t2 t1' t2',
      eqT t1 t1' ->
      eqT t2 t2' ->
      eqT (pairT t1 t2) (pairT t1' t2')
  | eqT_pair_eta : forall t,
      lclosedT t ->
      eqT (pairT (fstT t) (sndT t)) t
  | eqT_fst : forall t t',
      eqT t t' ->
      eqT (fstT t) (fstT t')
  | eqT_fst_beta : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      eqT (fstT (pairT t1 t2)) t1
  | eqT_snd : forall t t',
      eqT t t' ->
      eqT (sndT t) (sndT t')
  | eqT_snd_beta : forall t1 t2,
      lclosedT t1 ->
      lclosedT t2 ->
      eqT (sndT (pairT t1 t2)) t2
  | eqT_refl : forall t,
      lclosedT t ->
      eqT t t
  | eqT_sym : forall t t',
      eqT t' t ->
      eqT t t'
  | eqT_trans : forall t t' t'',
      eqT t t' ->
      eqT t' t'' ->
      eqT t t''
.

Hint Constructors eqT.

(*************************************************************************)
(* Typing *)
(*************************************************************************)

Inductive okG : tenv -> eenv -> Prop :=
  | okG_empty : forall D,
      okD D ->
      okG D empty
  | okG_var : forall D x t G,
      okG D G ->
      x \notin dom G ->
      ofT D t omK ->
      okG D (G & x~t)
.

Inductive ofE : tenv -> eenv -> exp -> typ -> Prop :=
  | ofE_var : forall D G (x : var) t,
      okG D G ->
      binds x t G ->
      ofE D G (varE x) t
  | ofE_triv : forall D G,
      okG D G ->
      ofE D G (trivE) unitT
  | ofE_lam : forall L D G t1 e2 t2,
      ofT D t1 omK ->
      (forall x : var, x \notin L -> ofE D (G & x~t1) (openXE x e2) t2) ->
      ofE D G (lamE t1 e2) (arrT t1 t2)
  | ofE_app : forall D G e1 e2 t2 t,
      ofE D G e1 (arrT t2 t) ->
      ofE D G e2 t2 ->
      ofE D G (appE e1 e2) t
  | ofE_pair : forall D G e1 e2 t1 t2,
      ofE D G e1 t1 ->
      ofE D G e2 t2 ->
      ofE D G (pairE e1 e2) (timT t1 t2)
  | ofE_fst : forall D G e t1 t2,
      ofE D G e (timT t1 t2) ->
      ofE D G (fstE e) t1
  | ofE_snd : forall D G e t1 t2,
      ofE D G e (timT t1 t2) ->
      ofE D G (sndE e) t2
  | ofE_lft : forall D G e1 t1 t2,
      ofE D G e1 t1 ->
      ofT D (plusT t1 t2) omK ->
      ofE D G (lftE e1 (plusT t1 t2)) (plusT t1 t2)
  | ofE_rgt : forall D G e2 t1 t2,
      ofE D G e2 t2 ->
      ofT D (plusT t1 t2) omK ->
      ofE D G (rgtE e2 (plusT t1 t2)) (plusT t1 t2)
  | ofE_roll : forall D G e k t,
      ofE D G e (openAT (recT k t) t) ->
      ofT D (recT k t) omK ->
      ofE D G (rollE e (recT k t)) (recT k t)
  | ofE_unroll : forall D G k t e,
      ofE D G e (recT k t) ->
      ofE D G (unrollE e) (openAT (recT k t) t)
  | ofE_gen : forall L D G k e t,
      okG D G ->
      (forall a : var, a \notin L ->
         ofE (D & a~k) G (openAE a e) (openAT a t)) ->
      ofE D G (genE k e) (allT k t)
  | ofE_inst : forall D G e1 k2 t t2,
      ofE D G e1 (allT k2 t) ->
      ofT D t2 k2 ->
      ofE D G (instE e1 t2) (openAT t2 t)
  | ofE_pack : forall D G t1 e2 k1 t2,
      ofT D t1 k1 ->
      ofE D G e2 (openAT t1 t2) ->
      ofT D (exT k1 t2) omK ->
      ofE D G (packE t1 e2 (exT k1 t2)) (exT k1 t2)
  | ofE_unpack : forall L D G e1 k1 t1 e2 t2,
      ofE D G e1 (exT k1 t1) ->
      (forall (a : var) (x : var), a \notin L -> x \notin L ->
         ofE (D & a~k1) (G & x~(openAT a t1)) (openXE x (openAE a e2)) t2) ->
      ofT D t2 omK ->
      ofE D G (unpackE e1 e2) t2
  | ofE_sub : forall D G e t' t,
      ofE D G e t' ->
      ofT D t omK ->
      eqT t' t ->
      ofE D G e t
.

Hint Constructors ofE.
