(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Omega.
Require Import List.

(*************************************************************************)
(* Some generic rewriting lemmas for naturals *)
(*************************************************************************)

Lemma plus_zero : forall n,
  n + 0 = n.
Proof.
  intros n. omega.
Qed.

Lemma zero_plus : forall n,
  0 + n = n.
Proof.
  intros n. omega.
Qed.

Lemma plus_one : forall n,
  n + 1 = S n.
Proof.
  intros n. omega.
Qed.

Lemma one_plus : forall n,
  1 + n = S n.
Proof.
  intros n. omega.
Qed.

Lemma plus_succ : forall n m,
  n + S m = S n + m.
Proof.
  intros n m. omega.
Qed.

Lemma plus_succ_l : forall n m,
  S n + m = S (n + m).
Proof.
  intros n m. omega.
Qed.

Lemma plus_succ_r : forall n m,
  n + S m = S (n + m).
Proof.
  intros n m. omega.
Qed.


(*************************************************************************)
(* Lemmas about lists *)
(*************************************************************************)

Lemma map_id : forall A (xs : list A),
  map (fun x => x) xs = xs.
Proof.
  intro A.
  induction xs; simpl map.
    reflexivity.
    f_equal. apply IHxs.
Qed.

Lemma map_eta : forall A B (f : A -> B),
  map (fun x => f x) = map f.
Proof.
  intros A B f. reflexivity.
Qed.

Lemma combine_refl : forall A (x x' : A) xs,
  In (x, x') (combine xs xs) -> x = x'.
Proof.
  intros A x x'.
  induction xs; simpl combine; intro Elem.
    inversion Elem.
    simpl In in Elem. destruct Elem.
      inversion H. subst. reflexivity.
      apply IHxs. apply H.
Qed.

Lemma in_combine : forall A B (xs : list A) (ys : list B) x y,
  In (x, y) (combine xs ys) ->
  In x xs /\ In y ys.
Proof.
  intros A B.
  induction xs; intros ys x y Elem; destruct ys; simpl combine in *; inversion Elem.
    inversion H. subst. split. apply in_eq. apply in_eq.
    split.
      apply in_cons. eapply proj1. apply IHxs. apply H.
      apply in_cons. eapply proj2. apply IHxs. apply H.
Qed.

Lemma in_combine_l_inv : forall A B (xs : list A) (ys : list B) x,
  length xs = length ys ->
  In x xs ->
  exists y,
  In (x, y) (combine xs ys).
Proof.
  intros A B xs ys x.
  generalize dependent ys.
  induction xs; intros ys Len Elem.
    inversion Elem.
    destruct ys; simpl length in Len; inversion Len.
      destruct Elem.
        exists b. subst. simpl combine. apply in_eq.
        destruct (IHxs ys H0 H). exists x0. simpl combine. apply in_cons. apply H1.
Qed.

Lemma in_combine_r_inv : forall A B (xs : list A) (ys : list B) y,
  length xs = length ys ->
  In y ys ->
  exists x,
  In (x, y) (combine xs ys).
Proof.
  intros A B xs ys y.
  generalize dependent xs.
  induction ys; intros xs Len Elem.
    inversion Elem.
    destruct xs; simpl length in Len; inversion Len.
      destruct Elem.
        exists a0. subst. simpl combine. apply in_eq.
        destruct (IHys xs H0 H). exists x. simpl combine. apply in_cons. apply H1.
Qed.

Lemma length_parts : forall A n1 n2 (xs : list A),
  length xs = n1 + n2 ->
  exists xs1, exists xs2,
  length xs1 = n1 /\ length xs2 = n2 /\ xs = xs1 ++ xs2.
Proof.
  intro A.
  induction n1; intros n2 xs Len; destruct xs as [ | x xs']; simpl length in *; simpl in Len.
    exists (nil (A := A)). exists (nil (A := A)).
      split. simpl length. reflexivity.
      split. simpl length. subst. reflexivity.
      simpl app. reflexivity.
    exists (nil (A := A)). exists (x::xs').
      split. simpl length. reflexivity.
      split. simpl length. apply Len.
      simpl app. reflexivity.
    inversion Len.
    inversion Len.
      destruct (IHn1 n2 xs' H0) as [xs1 [xs2 [Len1 [Len2 Concat]]]].
      exists (x::xs1). exists xs2.
        split. simpl length. f_equal. apply Len1.
        split. apply Len2.
        simpl app. rewrite Concat. reflexivity.
Qed.

Lemma length_zero_is_nil : forall A (Xs:list A) , length Xs = 0 -> Xs = nil.
Proof.
  intros A.
  induction Xs.
  intros. 
  reflexivity.
  intro.
  inversion H.
Qed.

Lemma length_succ_is_cons : forall n A (Xs:list A) , length Xs = S n -> exists Y, exists Ys, Xs = Y::Ys.
Proof.
  intros n A.
  induction Xs.
  intro.
  inversion H.
  intros.
  inversion H.
  exists a.
  exists Xs.
  reflexivity.
Qed.
